#!/bin/sh

abstract() {
  class "$1" "$2" "abstract"
}

class() {  
  # The classname is the string before the : on the first line
  classname=`echo "$1" | sed -n '1 s/ *:.*$//p'`
  # The parent's class name is the string after the : on the first line
  parent=`echo "$1" | sed -n '1 s/^.*: *//p'` ; # String after :
  # The fields come from the second line through the first empty line
  # Each is the identifier just before the semicolon
  fields=`echo "$1" | sed '/typedef/d' | sed -n '2,/^\$/ s/^.*[^a-zA-Z0-9_]\([a-zA-Z0-9_]*\);.*/\1/p'`
  # The body for the header file starts at the second line
  hppbody=`echo "$1" | sed -n '2,$p'`

  # Any additional methods are defined in the second argument

  #echo "[$classname]"
  #echo "[$parent]"
  #echo "[$fields]"
  #echo "[$hppbody]"

  forwarddefs="$forwarddefs
  class $classname;"

  # Define a default (zero-argument) constructor if one isn't already
  # defined in the body
  if (echo $hppbody | grep -q "$classname()"); then
    defaultconstructor=
  else
    defaultconstructor="$classname() {}
"    
  fi

  if test -z "$3"; then
      visitorclassdefs="$visitorclassdefs
  virtual Status visit($classname& n) { assert(0); return Status(); }"
      welcome="
    IRCLASSDEFS;
  public:
    Status welcome(Visitor&);"
      welcomedef="
  IRCLASS($classname);
  Status $classname::welcome(Visitor &v) { return v.visit(*this); }"    
  else
    welcome="public:"
    welcomedef=
  fi

  classdefs="$classdefs

  class $classname : public $parent {
    $welcome
    void read(XMListream &);
    void write(XMLostream &) const;
    $defaultconstructor
$hppbody
  };
"

  if test -n "$fields"; then
    writefields=`echo $fields | sed "s/ / << /g"`;
    writefields="
    w << $writefields;"
    readfields=`echo $fields | sed "s/ / >> /g"`;
    readfields="
    r >> $readfields;"
  else
    readfields=
    writefields=
  fi
  
  methoddefs="$methoddefs

  void $classname::read(XMListream &r) {
    $parent::read(r); $readfields
  }

  void $classname::write(XMLostream &w) const {
    $parent::write(w); $writefields
  }
$welcomedef
  $2
  "
}

abstract "ASTNode : Node


  virtual Status welcome(Visitor&) = 0;
"
abstract "Symbol : ASTNode
  string name;

  Symbol(string s) : name(s) {}"
class "ModuleSymbol : Symbol
  Module *module;

  ModuleSymbol(string s) : Symbol(s), module(0) {}"
class "SignalSymbol : Symbol
  TypeSymbol *type;
  string direction; // input, output, inputoutput, sensor, return, local
  FunctionSymbol *combine; // combining function, if any
  Expression *initializer;
  VariableSymbol *presence;
  VariableSymbol *value;  

  SignalSymbol(string n, TypeSymbol *t, string d, FunctionSymbol *f,
               Expression *e, VariableSymbol *p, VariableSymbol *v)
    : Symbol(n), type(t), direction(d), combine(f), initializer(e),
      presence(p), value(v) {}"
class "BuiltinSignalSymbol : SignalSymbol

  BuiltinSignalSymbol(string n, TypeSymbol *t, string d, FunctionSymbol *f,
                      VariableSymbol *p, VariableSymbol *v)
    : SignalSymbol(n, t, d, f, NULL, p, v) {}"
class "TrapSymbol : Symbol
  TypeSymbol *type;
  Expression *initializer;
  int code;
  VariableSymbol *presence;
  VariableSymbol *value;

  TrapSymbol(string s, TypeSymbol *t, Expression *e, VariableSymbol *p,
             VariableSymbol *v)
    : Symbol(s), type(t), initializer(e), code(0), presence(p), value(v) {}"
class "VariableSymbol : Symbol
  TypeSymbol *type;
  Expression *initializer;

  VariableSymbol(string n, TypeSymbol *t, Expression *e)
    : Symbol(n), type(t), initializer(e) {}"
class "ConstantSymbol : VariableSymbol

  ConstantSymbol(string n, TypeSymbol *t, Expression *i)
     : VariableSymbol(n, t, i) {}"
class "BuiltinConstantSymbol : ConstantSymbol

  BuiltinConstantSymbol(string n, TypeSymbol *t, Expression *i)
     : ConstantSymbol(n, t, i) {}"
class "TypeSymbol : Symbol

   TypeSymbol(string s) : Symbol(s) {}"
class "BuiltinTypeSymbol : TypeSymbol

   BuiltinTypeSymbol(string s) : TypeSymbol(s) {}"
class "FunctionSymbol : TypeSymbol
  vector<TypeSymbol*> arguments;
  TypeSymbol *result;

  FunctionSymbol(string s) : TypeSymbol(s), result(NULL) {}"
class "BuiltinFunctionSymbol : FunctionSymbol

  BuiltinFunctionSymbol(string s) : FunctionSymbol(s) {}"
class "ProcedureSymbol : TypeSymbol
  vector<TypeSymbol*> reference_arguments;
  vector<TypeSymbol*> value_arguments;

  ProcedureSymbol(string s) : TypeSymbol(s) {}"
class "TaskSymbol : ProcedureSymbol

  TaskSymbol(string s) : ProcedureSymbol(s) {}"
class "SymbolTable : ASTNode
  SymbolTable *parent;
  typedef map<string, Symbol*> stmap;
  stmap symbols;

  SymbolTable() : parent(NULL) {}

  class iterator {
    stmap::iterator i;
  public:
    iterator(stmap::iterator ii) : i(ii) {}
    void operator ++(int) { i++; } // int argument denotes postfix
    void operator ++() { ++i; } // int argument denotes postfix
    bool operator !=(const iterator &ii) { return i != ii.i; }
    Symbol *operator *() { return (*i).second; }
  };

  iterator begin() { return iterator(symbols.begin()); }
  iterator end() { return iterator(symbols.end()); }

  bool local_contains(const string) const;
  bool contains(const string) const;
  void enter(Symbol *);
  Symbol* get(const string);
" "
bool SymbolTable::local_contains(const string s) const {
  return symbols.find(s) != symbols.end();
}

bool SymbolTable::contains(const string s) const {
  for ( const SymbolTable *st = this ; st ; st = st->parent )
    if (st->symbols.find(s) != st->symbols.end()) return true;
  return false;
}

void SymbolTable::enter(Symbol *sym) {
  assert(sym);
  assert(symbols.find(sym->name) == symbols.end());
  symbols.insert( std::make_pair(sym->name, sym) );
}

Symbol* SymbolTable::get(const string s) {
  map<string, Symbol*>::const_iterator i;
  for ( SymbolTable *st = this; st ; st = st->parent ) {
    i = st->symbols.find(s);
    if (i != st->symbols.end()) {
      assert((*i).second);
      assert((*i).second->name == s);
      return (*i).second;
    }
  }
  assert(0); // get should not be called unless contains returned true
}
"
abstract "Expression : ASTNode
  TypeSymbol *type;

  Expression(TypeSymbol *t) : type(t) {}"

class "Literal : Expression
  string value;

  Literal(string v, TypeSymbol *t) : Expression(t), value(v) {}"
class "LoadVariableExpression : Expression
  VariableSymbol *variable;

  LoadVariableExpression(VariableSymbol *v)
    : Expression(v->type), variable(v) {}"
class "LoadSignalExpression : Expression
  SignalSymbol *signal;

  LoadSignalExpression(SignalSymbol *s)
    : Expression(s->type), signal(s) {}"
class "LoadSignalValueExpression : LoadSignalExpression
  
  LoadSignalValueExpression(SignalSymbol *s) : LoadSignalExpression(s) {}"
class "LoadTrapExpression : Expression
  TrapSymbol *trap;

  LoadTrapExpression(TrapSymbol *t)
   : Expression(t->type), trap(t) {}"
class "LoadTrapValueExpression : LoadTrapExpression
  
  LoadTrapValueExpression(TrapSymbol *s) : LoadTrapExpression(s) {}"

class "UnaryOp : Expression
  string op;
  Expression *source;

  UnaryOp(TypeSymbol *t, string s, Expression *e)
    : Expression(t), op(s), source(e) {}"
class "BinaryOp : Expression
  string op;
  Expression *source1;
  Expression *source2;

  BinaryOp(TypeSymbol *t, string s, Expression *e1, Expression *e2)
    : Expression(t), op(s), source1(e1), source2(e2) {}"  
class "FunctionCall : Expression
  FunctionSymbol *callee;
  vector<Expression*> arguments;

  FunctionCall(FunctionSymbol *s)
    : Expression(s->result), callee(s) {}"
class "Delay : Expression
  Expression *predicate;
  Expression *count;
  VariableSymbol *counter;
  bool is_immediate;

  Delay(TypeSymbol *t, Expression *e1, Expression *e2, VariableSymbol *v,
         bool i) : Expression(t), predicate(e1), count(e2), counter(v),
                   is_immediate(i) {}"

class "CheckCounter : Expression
  VariableSymbol *counter;

  CheckCounter(TypeSymbol *t, VariableSymbol *c): Expression(t), counter(c) {}
"
class "Module : ASTNode
  ModuleSymbol *symbol;
  SymbolTable *types;
  SymbolTable *constants;
  SymbolTable *functions;
  SymbolTable *procedures;
  SymbolTable *tasks;
  SymbolTable *signals;
  SymbolTable *variables;
  vector<InputRelation*> relations;
  ASTNode *body;
  int max_code;
  
  Module() : max_code(0) {}
  Module(ModuleSymbol *);
  ~Module();    
" "
Module::Module(ModuleSymbol *s) : symbol(s), body(NULL) {
  signals = new SymbolTable();
  constants = new SymbolTable();
  types = new SymbolTable();
  functions = new SymbolTable();
  procedures = new SymbolTable();
  tasks = new SymbolTable();
  variables = new SymbolTable();
}

Module::~Module() {
   delete signals;
   delete types;
   delete constants;
   delete functions;
   delete procedures;
   delete tasks;
   delete body;
   delete variables;
}"
abstract "InputRelation : ASTNode"

class "Exclusion : InputRelation
  vector<SignalSymbol *> signals;"

class "Implication : InputRelation
  SignalSymbol *predicate;
  SignalSymbol *implication;

  Implication(SignalSymbol *ss1, SignalSymbol*ss2)
    : predicate(ss1), implication(ss2) {}"
class "Modules : ASTNode
  SymbolTable module_symbols;
  vector<Module*> modules;

  void add(Module*);
" "
void Modules::add(Module* m) {
  assert(m);
  assert(m->symbol);
  assert(!module_symbols.contains(m->symbol->name));
  modules.push_back(m);
  module_symbols.enter(m->symbol);
}"
abstract "Statement : ASTNode"

abstract "BodyStatement : Statement
  Statement *body;

  BodyStatement(Statement *s) : body(s) {}"
class "PredicatedStatement : BodyStatement
  Expression *predicate;

  PredicatedStatement(Statement *s, Expression *e)
    : BodyStatement(s), predicate(e) {}"
abstract "CaseStatement : Statement
  vector<PredicatedStatement *> cases;
  Statement *default_stmt;

  CaseStatement() : default_stmt(0) {}
  PredicatedStatement *newCase(Statement *s, Expression *e) {
    PredicatedStatement *ps = new PredicatedStatement(s, e);
    cases.push_back(ps);
    return ps;
  }"

class "StatementList : Statement
  vector<Statement *> statements;

  StatementList& operator <<(Statement *s) {
    assert(s);
    statements.push_back(s);
    return *this;
  }"
class "ParallelStatementList : Statement 
  vector<Statement *> threads;"

class "Nothing : Statement"
class "Pause : Statement"
class "Halt : Statement"

class "StartCounter : Statement
  Expression *predicate;
  Expression *startvalue;
  VariableSymbol *counter;
  STNode *st;

  StartCounter(Expression *p, Expression *sv, VariableSymbol *v): predicate(p), startvalue(sv), counter(v) {}"

class "Emit : Statement
  SignalSymbol *signal;
  Expression *value;

  Emit(SignalSymbol *s, Expression *e) : signal(s), value(e) {}"

class "Sustain : Emit

  Sustain(SignalSymbol *s, Expression *e) : Emit(s, e) {}"

class "Assignment : Statement
  VariableSymbol *variable;
  Expression *value;

  Assignment(VariableSymbol *v, Expression *e) : variable(v), value(e) {}"
class "ProcedureCall : Statement
  ProcedureSymbol *procedure;
  vector<VariableSymbol*> reference_args;
  vector<Expression*> value_args;

  ProcedureCall(ProcedureSymbol *ps) : procedure(ps) {}"
class "Present : CaseStatement"
class "If : CaseStatement"
class "Loop : BodyStatement

  Loop(Statement *s) : BodyStatement(s) {}"

class "Repeat : Loop
  Expression *count;
  bool is_positive;

  Repeat(Statement *s, Expression *e, bool p)
    : Loop(s), count(e), is_positive(p) {}"
class "Abort : CaseStatement
  Statement *body;
  bool is_weak;
  int code;

  Abort(Statement *s, bool i) : body(s), is_weak(i), code(0) {}
  Abort(Statement *s, Expression *e, Statement *s1)
    : body(s), is_weak(false), code(0) {
    newCase(s1, e);
  }"

class "Await : CaseStatement"

class "LoopEach : PredicatedStatement

  LoopEach(Statement *s, Expression *e) : PredicatedStatement(s, e) {}"

class "Every : PredicatedStatement

  Every(Statement *s, Expression *e) : PredicatedStatement(s, e) {}"

class "Suspend : PredicatedStatement

  Suspend(Statement *s, Expression *e) : PredicatedStatement(s, e) {}"

class "DoWatching : PredicatedStatement
  Statement *timeout;

  DoWatching(Statement *s1, Expression *e, Statement *s2)
    : PredicatedStatement(s1, e), timeout(s2) {}"

class "DoUpto : PredicatedStatement

  DoUpto(Statement *s, Expression *e) : PredicatedStatement(s, e) {}"
class "TaskCall : ProcedureCall
  SignalSymbol *signal;
  Statement *body;

  TaskCall(TaskSymbol *ts) : ProcedureCall(ts), signal(0), body(0) {}
"

class "Exec : Statement
  vector <TaskCall *> calls;"
abstract "ScopeStatement : BodyStatement
   SymbolTable *symbols;"

class "Trap : ScopeStatement
  vector<PredicatedStatement *> handlers;

  PredicatedStatement* newHandler(Expression *e, Statement *s) {
    PredicatedStatement *ps = new PredicatedStatement(s, e);
    handlers.push_back(ps);
    return ps;
  }"

class "Exit : Statement
  TrapSymbol *trap;
  Expression *value;

  Exit(TrapSymbol *t, Expression *e) : trap(t), value(e) {}"

class "Signal : ScopeStatement"

class "Var : ScopeStatement"
abstract "Renaming : ASTNode
  string old_name;

Renaming(string s) : old_name(s) {}"

class "TypeRenaming : Renaming
  TypeSymbol *new_type;

  TypeRenaming(string s, TypeSymbol *t) : Renaming(s), new_type(t) {}"  

class "ConstantRenaming : Renaming
  Expression *new_value;

  ConstantRenaming(string s, Expression *e) : Renaming(s), new_value(e) {}"

class "FunctionRenaming : Renaming
  FunctionSymbol *new_func;

  FunctionRenaming(string s, FunctionSymbol *f) : Renaming(s), new_func(f) {}"

class "ProcedureRenaming : Renaming
  ProcedureSymbol *new_proc;

  ProcedureRenaming(string s, ProcedureSymbol *p)
    : Renaming(s), new_proc(p) {}"

class "SignalRenaming : Renaming
  SignalSymbol *new_sig;

  SignalRenaming(string s, SignalSymbol *ss) : Renaming(s), new_sig(ss) {}"

class "Run : Statement
  string old_name;
  string new_name;
  vector<TypeRenaming *> types;
  vector<ConstantRenaming *> constants;
  vector<FunctionRenaming *> functions;
  vector<ProcedureRenaming *> procedures;
  vector<ProcedureRenaming *> tasks;
  vector<SignalRenaming *> signals;

  Run(string s) : old_name(s), new_name(s) {}"
class "IfThenElse : Statement
  Expression *predicate;
  Statement *then_part;
  Statement *else_part;

  IfThenElse(Expression *e) : predicate(e) , then_part(0), else_part(0) {}
  IfThenElse(Expression *e, Statement *s1, Statement *s2)
    : predicate(e) , then_part(s1), else_part(s2) {}"
class "Goto : Statement
  Label *target;

  Goto(Label *s) : target(s) {}"
class "Label : Statement
  string name;

  Label(string s) : name(s) {}"
class "Catch : BodyStatement
  int code;
  Label *label;

  Catch(Statement *s, int c, Label *l) : BodyStatement(s), code(c), label(l) {}
"  
abstract "Handler : Statement
  vector<Catch*> catches;
  Label *catch0;

  Handler(Label *c0) : catch0(c0) {}
  void newCatch(Statement *s, int c, Label *l) {
    catches.push_back(new Catch(s, c, l));
  }
"
class "Try : Handler
  Statement *body;

  Try(Label *c0, Statement *s) : Handler(c0), body(s) {}
"
class "Break : Statement
  int level;
  Label *continue_target;
  Statement *set_state;

  Break(int l) : level(l), continue_target(0), set_state(0) {}"
abstract "Continuable : Handler
  Label *continue_label;

  Continuable(Label *c0, Label *cont) : Handler(c0), continue_label(cont) {}
"
class "Resume : Continuable
  Statement *body;
  Statement *branch;

  Resume(Label *c0, Label *cont, Statement *s)
    : Continuable(c0, cont), body(s), branch(0) {}
"
class "Continue : Statement"
class "Thread : Resume
  VariableSymbol *exit_level;

  Thread(Label *c0, Label *cont, Statement *b, VariableSymbol *el)
     : Resume(c0, cont, b), exit_level(el) {}
"
class "Parallel : Continuable
  vector<Thread*> threads;
  Statement *synchronize;

  Parallel(Label *c0, Label *cont)
    : Continuable(c0, cont), synchronize(0) {}
  void newThread(Label *c0, Label *cont, Statement *b, VariableSymbol *el) {
    threads.push_back(new Thread(c0, cont, b, el));
  }
"
abstract "GRCNode : ASTNode
  vector<GRCNode*> predecessors;
  vector<GRCNode*> successors;

  virtual Status welcome(Visitor&) = 0;

  GRCNode& operator >>(GRCNode*);
" "
  GRCNode& GRCNode::operator >>(GRCNode * s) {
    successors.push_back(s);
    if (s) s->predecessors.push_back(this);
    return *this;
  }
"
class "STNode : GRCNode"
class "STexcl : STNode"
class "STpar : STNode"
class "STref : STNode
  int type;

  STref(): type(0) {}

  int isabort() { return type == 1;}
  void setabort() { type = 1;}
  int issuspend() { return type == 2;}
  void setsuspend() { type = 2;}
"
class "STleaf : STNode
  int type;

  STleaf(): type(0) {}
 
  int isfinal() { return type == 1;}
  void setfinal() { type = 1;}
"
class "TopGRC : GRCNode"
class "Nop : GRCNode
  int type;
  int code;

  Nop(): type(0), code(0) {}

  int isflowin() { return type == 1;}
  void setflowin() { type = 1;}
  // a shorcut Nop gives "up" flow to child 0
  int isshortcut() { return type == 2;}
  void setshortcut() { type = 2;}
"
class "RecT1 : GRCNode"
class "DefineSignal : GRCNode
  SymbolTable *signals;

  DefineSignal(Signal &s) : signals(s.symbols){}
"
class "UndefineSignal : GRCNode
  SymbolTable *signals;

  UndefineSignal(Signal &s): signals(s.symbols) {}
"
class "Switch : GRCNode
  STNode *st;  
"
class "Test : GRCNode
  Expression *predicate;
  STNode *st;
  
  Test(Expression *e) : predicate(e) {}
"
class "Fork : GRCNode
"
class "Sync : GRCNode
  STNode *st;  
  int depth;
 
  Sync() : depth(0) {}

  void setdepth() { depth = 1; }
  int isdepth() { return depth == 1; }
"
class "Terminate : GRCNode
  int code;

  Terminate(int c) : code(c) {}
"
class "Action : GRCNode
  Statement *body;

  Action(Statement *s) : body(s) {}
"
class "Enter : GRCNode
  STNode *st;
"
class "Leave : GRCNode
  STNode *st;
"

######################################################################

echo "#ifndef _AST_HPP
#  define _AST_HPP

/* Automatically generated by AST.sh -- do not edit */

#  include \"IR.hpp\"
#  include <string>
#  include <vector>
#  include <map>

namespace AST {
  using IR::Node;
  using IR::XMListream;
  using IR::XMLostream;
  using std::string;
  using std::vector;
  using std::map;

  class Visitor;
$forwarddefs

  union Status {
    int i;
    ASTNode *n;
    Status() {}
    Status(int ii) : i(ii) {}
    Status(ASTNode *nn) : n(nn) {}
  };

$classdefs

  class Visitor {
  public:
$visitorclassdefs
  };

}

#endif
" > AST.hpp

echo "/* Automatically generated by AST.sh -- do not edit */
#include \"AST.hpp\"
namespace AST {

$methoddefs

}
" > AST.cpp
