#include "cverilog.hpp"

namespace CVER {



 // SR FLIP FLOP - S has priority over R

  sr_ff::sr_ff(wire *q, wire *s, wire *r):s(s),r(r),q(q)
  {
    assert(q && (s||r));
    rst=current->rst;
    clk=current->tick;
    init=0;
    name=current->new_uname(); current->modules.push_back(this);
  }

  void sr_ff::pr_def()
  {
    ostream &o=current->o;

    o<<"module sr_ff(rst, clk, q, s, r, init);\n";
    o<<"input rst, clk, s, r, init;\n";
    o<<"output q;\n";
    o<<"reg q;\n";

    o<<"always @(posedge clk or posedge rst) begin\n";
    o<<"  if(rst==1) q <= init; else\n";
    o<<"    if(s==1) q <= 1; else if(r==1) q<=0;\n";
    o<<"end\n";

    o<<"endmodule\n\n";
  }

  // s or r can be missing
  void sr_ff::pr_instance()
  {
    ostream &o=current->o;
    assert(rst&&clk&&q&&  (s||r));
    o<<"sr_ff "<<name<<"("<<rst->name<<","<<clk->name<<","
     <<q->name<<","<<wn(s,0)<<","<<wn(r,0)<<","<<init<<");\n";
  }

  void sr_ff::dw_dot()
  {
    ostream &o=current->odot;
    assert(rst&&clk&&q&& (s||r));
    o<<name<<"[label=\"SRFF\" shape=box ]\n";
    drawin(s,"s");
    drawin(r,"r");
    drawout(q,"q");
  }

  void sr_ff::req_wire(vector<wire*> &win, vector<wire*> &wout)
  {
    assert(s || r); assert(q);
    if(s) win.push_back(s);
    if(r) win.push_back(r);
    wout.push_back(q);
  }


  // D FLIP-FLOP

  d_ff::d_ff(wire *q, wire *d):d(d),q(q)
  {
    assert(q);
    //    assert(d);
    rst=current->rst;
    clk=current->tick;
    init=0;
    name=current->new_uname(); current->modules.push_back(this);
  }

  d_ff::d_ff(wire *q, wire *d, string pref, int ix):d(d), q(q)
  {
    stringstream ss;
    assert(q);
    // assert(d);
    rst=current->rst;
    clk=current->tick;
    init=0;
    ss<<pref<<ix;
    name=ss.str(); current->modules.push_back(this);
  }

  void d_ff::pr_def()
  {
    ostream &o=current->o;

    o<<"module d_ff(rst, clk, q, d, init);\n";
    o<<"input rst, clk, d, init;\n";
    o<<"output q;\n";
    o<<"reg q;\n";

    o<<"always @(posedge clk or posedge rst) begin\n";
    o<<"  if(rst==1) q <= init; else\n";
    o<<"      q <= d;\n";
    o<<"end\n";

    o<<"endmodule\n\n";
  }

  void d_ff::pr_instance()
  {
    ostream &o=current->o;
    assert(rst&&clk&&q);
    o<<"d_ff "<<name<<"("<<rst->name<<","<<clk->name<<","
     <<q->name<<","<<wn(d,0)<<","<<init<<");\n";
  }
 
  void d_ff::dw_dot()
  {
    ostream &o=current->odot;
    assert(rst&&clk&&q);
    o<<name<<"[label=\"DFF"<<(init?"*":"")<<"\\n"<<name<<"\" shape=box]\n";
    drawin(d,"d");
    drawout(q,"q");
  }

  void d_ff::req_wire(vector<wire*> &win, vector<wire*> &wout)
  {
    assert(d); assert(q);
    win.push_back(d);
    wout.push_back(q);
  }


  // AND GATE

  andgate::andgate(wire *out, wire *in1, wire *in2): out(out)
  {
    assert(out && in1 && in2);
    win.push_back(in1);
    win.push_back(in2);
    delay=current->delay;
    name=current->new_uname();
    current->modules.push_back(this);
  }

  andgate::andgate()
  {
    delay=current->delay;
    name=current->new_uname();
    current->modules.push_back(this);
  }

  void andgate::pr_instance()
  {
    ostream &o=current->o;
    assert(out);
    o<<"and #"<<delay<<" "<<name<<"("<<out->name;

    for(vector<wire*>::iterator w=win.begin(); w!=win.end(); w++){
      assert(*w);
      o<<","<<(*w)->name;
    }
    o<<");\n";
  }

  void andgate::dw_dot()
  {
    ostream &o=current->odot;
    o<<name<<"[label=\"\" shape=box style=filled color=red height=0.2 width=0.3]\n";
    drawin(win);
    drawout(out,"");
  }

   void andgate::req_wire(vector<wire*> &win, vector<wire*> &wout)
   {
     assert(out);
     win = this->win;
     wout.push_back(out);
   }

  // OR GATE

  orgate::orgate(wire *out, wire *in1, wire *in2): out(out)
  {
    assert(out && in1 && in2);
    win.push_back(in1);
    win.push_back(in2);
    delay=current->delay;
    name=current->new_uname();
    current->modules.push_back(this);
  }

  orgate::orgate()
  {
    delay=current->delay;
    name=current->new_uname();
    current->modules.push_back(this);
  }

  void orgate::pr_instance()
  {
    ostream &o=current->o;
    assert(out);
    o<<"or #"<<delay<<" "<<name<<"("<<out->name;

      for(vector<wire*>::iterator w=win.begin(); w!=win.end(); w++){
	assert(*w);
	o<<","<<(*w)->name;
      }
    o<<");\n";

  }

  void orgate::dw_dot()
  {
    ostream &o=current->odot;
    o<<name<<"[label=\"\" shape=box style=filled color=blue height=0.2 width=0.3 ]\n";
    drawin(win);
    drawout(out,"");
  }
  
  void orgate::req_wire(vector<wire*> &win, vector<wire*> &wout)
  {
    assert(out);
    win = this->win;
    wout.push_back(out);
  }



  // NOT GATE

  notgate::notgate(wire *out, wire *in): out(out),in(in)
  {
    assert(out && in);
    delay=current->delay;
    name=current->new_uname();
    current->modules.push_back(this);
  }

  void notgate::pr_instance()
  {
    ostream &o=current->o;
    assert(out&&in);
    o<<"not #"<<delay<<" "<<name<<"("<<out->name<<","<<in->name<<");\n";
  }

  void notgate::dw_dot()
  {
    ostream &o=current->odot;
    assert(in && out);
    o<<name<<"[label=\"\" style=filled shape=circle color=black height=0.2 width=0.2]\n";
    drawin(in,"");
    drawout(out,"");
  }  

  void notgate::req_wire(vector<wire*> &win, vector<wire*> &wout)
  {
    assert(in && out);
    win.push_back(in);
    wout.push_back(out);
  }

  counter::counter(wire *alarm, wire *start, wire *cenable, int startval):
    alarm(alarm),start(start),cenable(cenable),startval(startval)
  {
    assert(alarm && cenable && start);
    delay=current->delay;
    name=current->new_uname();
    current->modules.push_back(this);    
  }

  void counter::pr_def()
  {
    ostream &o=current->o;

    o<<"module counter(rst, clk, alarm, start, enable, startval);\n";
    o<<"input rst, clk,start, enable, startval;\n";
    o<<"output alarm;\n";
    
    o<<"wire [15:0] startval, d, nxd;\n";
    o<<"reg [15:0] q;\n";
    o<<"wire alarm;\n";
 
    o<<"assign nxd = (q==0 || ! enable) ? q : q-1;\n";
    o<<"assign d = start ? startval : nxd;\n";
    o<<"assign alarm = (q == 0 && enable);\n";
    
    o<<"always @(posedge clk or posedge rst) begin\n";
    o<<"  if(rst==1) q = 0; else q <= d;\n";
    o<<"end\n";
    
    o<<"endmodule\n";
    
  }

  void counter::pr_instance()
  {
    ostream &o=current->o;
    assert(alarm && start && cenable);
    o<<"counter "<<name<<"(rst,tick,"<<alarm->name<<","<<wn(start,0)<<","<<
      wn(cenable,0)<<","<<startval<<");\n";
  }
  

  void counter::dw_dot()
  {
    ostream &o=current->odot;
    assert(alarm && cenable && start);
    o<<name<<"[label=\""<<startval<<"\" style=filled shape=box fillcolor=yellow]\n";
    drawout(alarm,"a");
    drawin(start,"s");
    drawin(cenable,"e");
  }
  
  void counter::req_wire(vector<wire*> &win, vector<wire*> &wout)
  {
    assert(alarm && cenable && start);
    win.push_back(cenable);
    win.push_back(start);
    wout.push_back(alarm);
  }


}


