#ifndef _GRC2ZAPHOD_
#define _GRC2ZAPHOD_

#include "IR.hpp"
#include "AST.hpp"
#include "GRCPrinter.hpp"
#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <set>

#include "cverilog.hpp"

using namespace CVER;
using namespace AST;

namespace ZAPHOD {

  extern int debug, sdebug, edebug;

  // signal header

  struct zaphod_signal {
    wire *signal;
    vector<wire*> sigemit;
    string realname;
  };

  typedef map<SignalSymbol*, zaphod_signal*> sigview;

  class GRC2zaphod;

// CONTEXT
  struct zaphod_context {
    GRC2zaphod *owner;   
    Module *mod;
    BuiltinTypeSymbol *int_type, *bool_type;

    TypeSymbol *t;
    int isconst;

    // bottom -> top
    wire *wexpr;
    int intexpr;

    // top -> bottom
    wire *flow;
    sigview *local_sig;

    zaphod_context(GRC2zaphod *z, Module *m): owner(z), mod(m) {
      int_type = dynamic_cast<BuiltinTypeSymbol*>(mod->types->get("integer"));
      bool_type = dynamic_cast<BuiltinTypeSymbol*>(mod->types->get("boolean"));
      assert( int_type && bool_type);
    }

    // these are "macros" for simple expression which accept only one type

    void eval_expr(Expression *e); // defined in grc2zaphod_utils.cpp    

    void setbool(wire *w){
      t = bool_type;
      wexpr = w;
    }

    wire* getbool(Expression *e){
      eval_expr(e);
      return wexpr;
    }

    void setint(int n, int cnst){
      t = int_type;
      intexpr = n;
      isconst = cnst;
    }

    int getint(Expression *e){
      eval_expr(e);
      assert( t == int_type );
      return intexpr;
    }

    int getintconst(Expression *e){
      eval_expr(e);
      assert(t == int_type);
      assert(isconst);
      return intexpr;
    }

    void do_statement(Statement *s, wire *flowin);
    wire *getflow() {return flow;}

  };

  // STATEMACHINE header

  class statemachine {

    typedef map<Sync *, vector<wire*> > trapmap;

    vector<wire*> enters;
    vector<wire*> leaves;
    vector<wire*> checks;
    vector<int> ff_type;
    vector<trapmap> enter_w;
    vector<trapmap> leave_w;
    
    wire *active;

    int id;
    static const int typeboot=0;
    static const int typeleaf=1;
    static const int typefin=2;
    static const int typepass=3;
    static const int typecache=4;
    static const int typeornet=5;

    int nostates, fststate;

    STNode *owner;
    GRC2zaphod *zaphod;

    void dump();

  public:

    statemachine(STNode *n, GRC2zaphod *z);

    void addenter(int no, wire *w, Sync *sy);
    wire *getstate(int no);
    wire *getactive();

    wire* wire_from_trapmap(trapmap &tm, int fftype);
    void sm2circuit(wire *susp);

  };

  // ZAPHODCOUNTER header
  struct zaphod_counter{
    vector<wire *> start;
    int startval; // hope they are the same !
    wire* enable; // the predicate to count
    wire* check; // the alarm
    STNode *st;

    wire* get_check();// creates the counter if necessary and rets the check wire

    zaphod_counter(): startval(0), enable(NULL), check(NULL) {}
  };


  // GRC2ZAPHOD header


  class GRC2zaphod: public Visitor {
    std::ostream &over, &odot;
    
  public:
    
    Module *mod;
    TopGRC *top;
    SymbolTable *sigs;
    STNode *stroot; Switch *top_sw;
    STleaf *boot;
    
    vector<GRCNode*> topo;
    std::set<GRCNode*> dfs_black;
    map<GRCNode*, Sync*> next_sync;
    map<Sync*, wire *> local_enters_valid;
    map<Sync*, wire *> fin_loop_valid;

    map<STNode*, wire*> suspend;

    map<VariableSymbol*, zaphod_counter> counters;

    string action;
    Cverilog cv;

    map<STNode*, statemachine*> machines;

    map<GRCNode*, map<GRCNode* , wire*> > cflow;

    map<SignalSymbol*, zaphod_signal> signals;
    sigview flat_local_signals, depth_local_signals;
    map<GRCNode*, sigview*> localsignals;

    Nodemap &refmap;

    zaphod_context zc;

    GRC2zaphod(std::ostream &over, std::ostream &odot, Module *mod, Nodemap &refmap, string action): over(over), odot(odot), mod(mod), action(action), cv(over, odot), refmap(refmap), zc(this, mod)  {

      top = dynamic_cast<AST::TopGRC*>(mod->body);
      assert(top);
      sigs = dynamic_cast<AST::SymbolTable*>(mod->signals);
      assert(sigs);
    }
    virtual ~GRC2zaphod() {}

    void visit (GRCNode *) {}

    Status visit(Switch &);
    Status visit(Test &);
    Status visit(Terminate &);
    Status visit(Sync &);
    Status visit(Fork &);
    Status visit(Action &);
    Status visit(Enter &);
    Status visit(Leave &);

    Status visit(TopGRC &);
    Status visit(STexcl &){return Status();}
    Status visit(STref &){return Status();}
    Status visit(STpar &){return Status();}
    Status visit(STleaf &){return Status();}

    Status visit(Nop &);
    Status visit(RecT1 &);

    Status visit(DefineSignal &);
    Status visit(UndefineSignal &) { assert(0);}

    // AST predicates & actions
    
    Status visit(LoadSignalExpression &);
    Status visit(BinaryOp &);
    Status visit(UnaryOp &);
    Status visit(LoadVariableExpression &);
    Status visit(VariableSymbol &s);
    Status visit(Literal &);
    Status visit(Assignment &);

    Status visit(StartCounter &);
    Status visit(CheckCounter &);
    Status visit(Emit &);

    Status visit(Delay &) { assert(0);}

    void children_sameflow(GRCNode *n, wire *flow);
    wire *flow_pair(GRCNode *p, GRCNode *c);
    wire *flow_fromparents(GRCNode *n);
    wire *dotest(GRCNode *n, wire* flowin, wire *wtest);

    void init();
    void execute();

    void dfs(GRCNode *n);
    void dfs_signals(GRCNode *n, int step, sigview* cview);
    void link_syncs();

    void setup_sm(STNode *n);
    void build_sm(STNode *n, wire* susp);

    void build_counters();

    void print_wirevec(string comment, vector<wire*> &wv);

  };

}

#endif
