#include <iostream>
#include <fstream>
#include <set>

#include "IR.hpp"
#include "AST.hpp"
#include "GRCPrinter.hpp"
#include "GRCWalker.hpp"


#define OPT_PRUNE_SW 1
#define OPT_MERGE_SW 2
#define OPT_RMV_EL 3
#define OPT_SHCUT 4
#define OPT_NULLP 10

namespace AST {

  int debug, sdebug;

  class GRCopt: public Visitor {

  public:
   
    TopGRC *top;
    SymbolTable *sigs;
    STNode *stroot;
    STleaf *boot;
    Switch *top_sw;
    int optstep;

    vector<GRCNode*> topo;
    std::set<GRCNode*> dfs_black;
    std::map<STNode*, int> stmark;

    std::set<GRCNode*> reachable;

    Nodemap &refmap;

    GRCopt(TopGRC *top, SymbolTable *sigs, Nodemap &refmap): top(top), sigs(sigs), refmap(refmap)  {}
    virtual ~GRCopt() {}

    void visit (GRCNode *) {}

    Status visit(Switch &);
    Status visit(Test &);
    Status visit(Terminate &);
    Status visit(Sync &);
    Status visit(Fork &);
    Status visit(Action &);
    Status visit(Enter &);
    Status visit(Leave &);

    Status visit(TopGRC &){return Status();}
    Status visit(STexcl &){return Status();}
    Status visit(STref &){return Status();}
    Status visit(STpar &){return Status();}
    Status visit(STleaf &){return Status();}

    Status visit(Nop &);
    Status visit(RecT1 &){return Status();}
    Status visit(DefineSignal &){ return Status();}
    Status visit(UndefineSignal &){ return Status();}

    vector<GRCNode*>::iterator findnode (vector<GRCNode*> &nvec,  GRCNode *n);

    void dostep(int step, int clear_black, int forward);
    void execute();

    void dfs(GRCNode *n);


    Status mergeSwitch (Switch &);
    Status pruneSwitch (Switch &);
    void rmv_simple (GRCNode *n);

    STNode* check_st(STNode *n, STNode *realpar);

  };

  Status GRCopt::visit(Test &){return Status();}
  Status GRCopt::visit(Terminate &){return Status();}
  Status GRCopt::visit(Sync &){return Status();}
  Status GRCopt::visit(Fork &){return Status();}
  Status GRCopt::visit(Action &){return Status();}
 
    
  void GRCopt::dfs(GRCNode *n)
  {
    int i;
    
    if(!n) return;

    if(dfs_black.count(n)>0) return;
    dfs_black.insert(n); // color black
    for(i=0;i<(int)n->successors.size();i++)
      dfs(n->successors[i]);
    topo.push_back(n);
  }

  STNode* GRCopt::check_st(STNode *n, STNode *realpar)
  {
    int keep;
    STNode* c;
    STref *ref;
    int is_simpleref, is_abort, is_suspend;

    n->predecessors[0]=realpar;

    if(dynamic_cast<STleaf*>(n)) {
      if(reachable.count(n) > 0){ keep=1; stmark[n]=1; return n; }
      keep=0; stmark[n]=0; return NULL;
    }

    is_simpleref = is_abort = 0;
    ref=dynamic_cast<STref*>(n);
    if(ref) {
      assert((int) n->successors.size()<=1);
      is_abort = ref->isabort();
      is_suspend = ref->issuspend();
      is_simpleref = ! (is_abort || is_suspend); 
    }

    if(is_simpleref) realpar = realpar; // dont keep anyawy
    else realpar = n; // may keep is any child is good

    c=NULL;
    keep=0;
    for(vector<GRCNode*>::iterator i=n->successors.begin(); i!=n->successors.end(); i++)
      if(*i){
	c = check_st((STNode*)(*i), realpar);
	(*i)=c;
	if(c) keep=1;
      }
    
    if(is_simpleref) {
      stmark[n]=0;
      if(keep) return c; else return NULL;
    }
    
    stmark[n]=keep;
    if(keep) return n;
    return NULL;
    
  }

  void GRCopt::dostep(int step, int clear_black, int forward)
  {
    int i, tsz;
    vector<GRCNode*>::iterator ip;
    GRCNode *n;

    optstep=step;

    topo.clear();
    if(clear_black) dfs_black.clear();
    dfs(top_sw);

    tsz=topo.size();

    for(i=tsz-1; i>=0; i--){
      n=topo[i];
      if(n==top_sw) continue;
      ip=n->predecessors.begin();
      while(ip!=n->predecessors.end()){
	if((!*ip) || dfs_black.count(*ip)==0) {
	  std::cerr<<refmap[n]<<"dfs: parent "<<refmap[*ip]<<" not in dfs_black - erase\n";
	  ip=n->predecessors.erase(ip);
	} else ip++;
      }
    }
    
    if(step==OPT_NULLP) return;

    if (forward)
      for(i=tsz-1; i>=0; i--){
	if(debug) std::cerr<<"Node "<<refmap[topo[i]]<<"\n";
	topo[i]->welcome(*this);    
      }
    else
      for(i = 0; i < tsz; i++) {
	if(debug) std::cerr<<"Node "<<refmap[topo[i]]<<"\n";
	topo[i]->welcome(*this);  
      }
    
  }

  void GRCopt::execute()
  {
    debug=0;
    sdebug=1;
    

    top_sw = dynamic_cast<Switch *>(top->successors[1]);
    stroot = dynamic_cast<STexcl *>(top->successors[0]);
    boot = dynamic_cast<STleaf* >(stroot->successors.back());
    assert(top_sw);
    assert(stroot);

    dostep(OPT_NULLP, 1, 1);

    std::cerr <<"Starting GRCWalker\n";

    GRCWalker gw(top, stroot, top_sw, boot, refmap, reachable);
    gw.execute();

    std::cerr <<"GRCWalker ok\n";

    /*

    int i, tsz;
    dfs_black.clear();
    dfs(stroot);
    tsz = topo.size();
    for(i = 0; i < tsz; i++) {
      if (reachable.count(topo[i]) == 0){
	if(dynamic_cast<STNode*>(topo[i]))
	  std::cerr<<"STNode"; else std::cerr<<"Flow node";
	std::cerr<<" "<<refmap[topo[i]]<<" is unreachable! \n";
      }
    }

    */

    check_st(stroot, (STNode*)top); // prunes the ST

    std::cerr<<"Pruning Switches\n";
    dostep(OPT_PRUNE_SW, 1, 0);
  
    std::cerr<<"Merging switches\n";
    dostep (OPT_MERGE_SW, 1, 1);

    std::cerr<<"Removing Leaves and Dangling Enters\n";
    dfs_black.clear();
    dfs(stroot);
    dostep(OPT_RMV_EL, 0, 1); // don't clear black, it contains st reacheable info

    std::cerr<<"Solving shortcut flow\n";
    //    dostep(OPT_SHCUT, 1, 1);

    dostep(OPT_NULLP, 1, 0);

  }

 vector<GRCNode*>::iterator GRCopt::findnode (vector<GRCNode*> &nvec, GRCNode *n)
  {

    vector<GRCNode*>::iterator i;

    for(i=nvec.begin(); i!=nvec.end(); i++)
      if((*i)==n) return i;

    assert(0);
    return i;
  }

  // can have several parents
  // links all parents to child
  // can be called for STNode, too

  void GRCopt::rmv_simple(GRCNode *n) 
  {
    GRCNode *c;
    vector<GRCNode*>::iterator i, ic, ip;

    if(n->successors.size() == 1){
      c=n->successors[0];
      assert(c);
      ic=findnode(c->predecessors, n);
      ic=c->predecessors.erase(ic);
    } else {
      c = NULL;
      if(sdebug) std::cerr<<"rmv_simple: stange or not, no successors\n";
    }

    for(i=n->predecessors.begin(); i!=n->predecessors.end(); i++){
      if(!*i) continue;
      ip=findnode((*i)->successors, n);
      if(c){
	(*ip) = c;
	ic = c->predecessors.insert(ic, *i);
	ic++;
      } else {
	assert( ! dynamic_cast<Switch*>(*i));
	(*i)->successors.erase(ip);
      }
    }

  }


  // SWITCH METHODS

  Status GRCopt::mergeSwitch (Switch &s)
  {

    Switch *p;
    int ich, szc, chno;
    GRCNode *ch, *st_par;
    vector<GRCNode*>::iterator i, ip;

   
    //@@@@
    //    assert(s.predecessors.size()==1);
    if (s.predecessors.size()>1)
      return Status();

    szc=s.successors.size();
    assert(szc == (int)s.st->successors.size());

    if((p=dynamic_cast<Switch*>(s.predecessors[0]))==NULL) return Status();

    ip=findnode(p->successors, &s);
    chno = ip-p->successors.begin();
    ip=p->successors.erase(ip);

    for(ich=0; ich<szc; ich++){
      ch=s.successors[ich];
      i=findnode(ch->predecessors, &s);
      (*i)=p;
      ip = p->successors.insert(ip, ch);
      ip++;
    }

    assert(s.st->predecessors.size() == 1);
    
    st_par = p->st;
    ip = chno+st_par->successors.begin();
    ip = st_par->successors.erase(ip);

    for (ich=0; ich<szc; ich++){
      ch = s.st->successors[ich];
      i = findnode(ch->predecessors, s.st);
      (*i) = st_par;
      ip = st_par->successors.insert(ip, ch);
      ip++;
    }

    return Status();

  }

  // finds NULL nodes in the ST and removes the branches

  Status GRCopt::pruneSwitch(Switch &s)
  {
    int nch;
    vector<GRCNode*>::iterator ch, sch, i;

    sch=s.st->successors.begin();
    while(sch != s.st->successors.end()){
      if(*sch && stmark[(STNode*)(*sch)]) {sch++; continue;} // continue if REAL children

      nch=sch - s.st->successors.begin();
      if(debug)std::cerr<<"removing child #"<<nch<<'\n';

      ch=s.successors.begin() + nch;
      std::cerr<<"grc search "<<refmap[*ch]<<'\n';
      i=findnode((*ch)->predecessors, &s);
      (*ch)->predecessors.erase(i);
      s.successors.erase(ch);

      if(*sch){
	std::cerr<<"st search\n";
	i=findnode((*sch)->predecessors, s.st);
	(*sch)->predecessors.erase(i);
      }
      sch = s.st->successors.erase(sch); // now sch points to the next element
    }

    /*
    if(s.successors.size() == 1 ){
      rmv_simple(&s);
      rmv_simple(s.st);
    }
    */
    

    return Status();
  }

  Status GRCopt::visit(Switch &s)
  {

    if (optstep == OPT_MERGE_SW)
      return mergeSwitch(s);

    if(optstep == OPT_PRUNE_SW)
      return pruneSwitch(s);


    return Status();
   
  }

  Status GRCopt::visit(Enter &s)
  {

    if (optstep != OPT_RMV_EL) return Status();

    if (dfs_black.count(s.st) == 0){
      std::cerr<<"Dangling Enter: "<<refmap[&s]<<'\n';
      rmv_simple(&s);
    }

    return Status();
  }

  Status GRCopt::visit(Leave &s)
  {
    if (optstep != OPT_RMV_EL)
      return Status();

    /*    if (dfs_black.count(s.st) == 0){
      std::cerr<<"Dangling leave: "<<refmap[&s]<<'\n';
      rmv_simple(&s);
    }
    */

    // remove anyway
    rmv_simple(&s);

    return Status();
  }

  Status GRCopt::visit(Nop &s)
  {
    Switch *sw;
    GRCNode *parsw, *ch0, *ch1;
    Nop *up;

    if(optstep != OPT_SHCUT || ! s.isshortcut()) return Status();

    assert(s.successors.size() == 2);
    ch0 = s.successors[0];
    ch1 = s.successors[1];

    assert(s.predecessors.size() == 1);
    sw = dynamic_cast<Switch*>(s.predecessors[0]);
    assert(sw);
    assert(sw->predecessors.size() == 1);
    parsw = sw->predecessors[0];
    assert(parsw);
    
    //insert up Nop between switch and its father
    up = new Nop();
    up->predecessors.push_back(parsw);
    up->successors.push_back(sw);
    (*findnode(parsw->successors, sw)) = up;
    sw->predecessors[0] = up;
    // ad child 0 (shcut) to up
    (*findnode(ch0->predecessors, &s)) = up;
    up->successors.push_back(ch0);

    // now remove ourselves
    (*findnode(sw->successors, &s)) = ch1;
    (*findnode(ch1->predecessors, &s)) = sw;

    return Status();
  }

}


int main(int argc, char *argv[])
{
  if(argc<5) {std::cerr<<"Usage: ... grc_file grc_opt_file dot_file dot_opt_file\n"; return -1;}

  std::ifstream grcfile(argv[1]); if(!grcfile.good()) return -1;
  std::ofstream grcoptfile(argv[2]); if(!grcoptfile.good()) return -1;
  
  std::ofstream dotfile(argv[3]); if(!dotfile.good()) return -1;
  std::ofstream dotoptfile(argv[4]); if(!dotoptfile.good()) return -1;
  


  try {

    IR::XMListream r(grcfile);
    IR::Node *n;
    r >> n;
    grcfile.close();
    
    AST::Modules *mods = dynamic_cast<AST::Modules*>(n);
   
    assert(mods->modules.size()==1);

    for ( std::vector<AST::Module*>::iterator i = mods->modules.begin() ;
	  i != mods->modules.end() ; i++ ) {
      assert(*i);

      AST::TopGRC *gn = dynamic_cast<AST::TopGRC*>((*i)->body);
      assert(gn);
      AST::SymbolTable *sig = dynamic_cast<AST::SymbolTable*>((*i)->signals);
      assert(sig);

      std::cerr << "File read - OK"<<std::endl;

      AST::Nodemap refmap;

      AST::GRCDot(dotfile,(AST::GRCNode *)gn,(*i),refmap); dotfile.close();

      AST::GRCopt grcopt(gn, sig, refmap);
      grcopt.execute();

      AST::GRCDot(dotoptfile, (AST::GRCNode *)gn,(*i),refmap); dotoptfile.close();

    }

    IR::XMLostream o(grcoptfile);
    o << n;
    grcoptfile.close();
    
  } catch (IR::Error &e) {
    std::cerr << e.s << std::endl;
    exit(-1);
  }

  return 0;
}
